/*
  rdesktop: A Remote Desktop Protocol client.
  Amiga front-end etc
  Copyright (C) Michal Zukowski 2012-2016
  Copyright (C) Martin Blom 2004-2005

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "rdesktop.h"
#include "config.h"
#include <dos/dos.h>
#include <workbench/startup.h>
#include <workbench/workbench.h>

#include <proto/asl.h>
#include <proto/cybergraphics.h>
#include <proto/dos.h>
#include <proto/exec.h>
#include <proto/graphics.h>
#include <proto/icon.h>
#include <proto/iffparse.h>
#include <proto/intuition.h>
#include <proto/keymap.h>
#include <proto/layers.h>
#include <proto/utility.h>
#include <proto/wb.h>
#include <proto/charsets.h>

#include <ctype.h>		/* isalpha */
#include <stdarg.h>		/* va_list va_start va_end */
#include <unistd.h>		/* read close getuid getgid getpid getppid gethostname */
#include <pwd.h>		/* getpwuid */
#include <sys/signal.h>		/* SIGINT */
#include "amiga_getkeylayout.h"

extern int amiga_left;
extern int amiga_top;
extern STRPTR amiga_pubscreen_name;
extern ULONG amiga_screen_id;
extern struct DiskObject* amiga_icon;

extern char g_title[64];
extern char g_username[64];
extern char g_hostname[16];
extern char g_keymapname[PATH_MAX];
extern int g_keylayout;
extern int g_width;
extern int g_height;
extern int g_tcp_port_rdp;
extern int g_server_depth;
extern int g_win_button_size;
extern RD_BOOL g_bitmap_compression;
extern RD_BOOL g_sendmotion;
extern RD_BOOL g_encryption;
extern RD_BOOL g_packet_encryption;
extern RD_BOOL g_desktop_save;
extern RD_BOOL g_polygon_ellipse_orders;
extern RD_BOOL g_fullscreen;
extern RD_BOOL g_clonescreen;
extern RD_BOOL g_grab_keyboard;
extern RD_BOOL g_hide_decorations;
extern RD_BOOL g_use_rdp5;
extern RD_BOOL g_console_session;
extern RD_BOOL g_numlock_sync;
extern RD_BOOL g_owncolmap;
extern RD_BOOL g_ownbackstore;
extern uint32 g_embed_wnd;
extern uint32 g_rdp5_performanceflags;
/* Session Directory redirection */
extern RD_BOOL g_redirect;
extern char g_redirect_server[64];
extern char g_redirect_domain[16];
extern char g_redirect_password[64];
extern char g_redirect_username[64];
extern char g_redirect_cookie[128];
extern uint32 g_redirect_flags;
extern RD_BOOL g_bitmap_cache_persist_enable;
extern RDPDR_DEVICE g_rdpdr_device[];
extern uint32 g_num_devices;
extern char *g_rdpdr_clientname;

uint32 ext_disc_reason = 0;

static void show_disconnect_reason(uint16 reason);

const char version[] = "$VER: " PACKAGE_STRING 
							  "(C) 2013-2016 Michal Zukowski, "
                              "(C) 2005-2007 Joerg Strohmayer, "
                              "(C) 2001-2008 Martin Blom, "
                              "(C) 1999-2005 Matthew Chapman et al.";
int __stack = 1024*100;

extern RD_BOOL g_rdpsnd;
extern int  amiga_audio_unit;

static const char template[] =
"SERVER/A,USER/K,DOMAIN/K,PASSWORD/K,CLIENT/K,CONSOLE/S,FRENCH/S,"
"NOENC=NOENCRYPTION/S,SHELL/K,DIR=DIRECTORY/K,LEFT/K/N,TOP/K/N,"
"W=WIDTH/K/N,H=HEIGHT/K/N,D=DEPTH/K/N,PUBSCREEN/K,TITLE/K,"
"FS=FULLSCREEN/S,CS=CLONESCREEN/S,SCREENMODE/K/N,AUDIO/K/N,REMOTEAUDIO/S,"
"NOMOUSE/S,EXP=EXPERIENCE/K,RDP4/S,KEYMAP/K/N,RDPC=RDPCOMPRESSION/S,"
"PC=PERSISTENTCACHE/S,SM=SLOWMOUSE/S,NETDISK/K";

static const char wbtemplate[] =
"SERVER/A/K,USER/K,DOMAIN/K,PASSWORD/K,CLIENT/K,CONSOLE/S,FRENCH/S,"
"NOENC=NOENCRYPTION/S,SHELL/K,DIR=DIRECTORY/K,LEFT/K/N,TOP/K/N,"
"W=WIDTH/K/N,H=HEIGHT/K/N,D=DEPTH/K/N,PUBSCREEN/K,TITLE/K,"
"FS=FULLSCREEN/S,CS=CLONESCREEN/S,SCREENMODE/K/N,AUDIO/K/N,REMOTEAUDIO/S,"
"NOMOUSE/S,EXP=EXPERIENCE/K,RDP4/S,KEYMAP/K/N,RDPC=RDPCOMPRESSION/S,"
"PC=PERSISTENTCACHE/S,SM=SLOWMOUSE/S,NETDISK/K,IGNORE/F";

static LONG  def_size  = 0;
static LONG  def_depth = 0;

static struct
{
    STRPTR a_server;
    STRPTR a_user;
    STRPTR a_domain;
    STRPTR a_password;
    STRPTR a_client;
    ULONG  a_console;
    ULONG  a_french;
    ULONG  a_noenc;
    STRPTR a_shell;
    STRPTR a_dir;
    ULONG* a_left;
    ULONG* a_top;
    LONG*  a_width;
    LONG*  a_height;
    ULONG* a_depth;
    STRPTR a_pubscreen;
    STRPTR a_title;
    ULONG  a_fullscreen;
	ULONG  a_clonescreen;
    ULONG* a_screenmode;
    ULONG* a_audio;
    ULONG  a_remoteaudio;
    ULONG  a_nomouse;
    STRPTR a_experience;
    ULONG  a_rdp4;
    ULONG* a_keymap;
    ULONG  a_rdp_compression;
    ULONG  a_persist_cache;
    ULONG  a_slowmouse;
    STRPTR a_netdisk;
    STRPTR a_ignore;
} a_args =
{
   NULL,
   NULL,
   NULL,
   NULL,
   NULL,
   FALSE,
   FALSE,
   FALSE,
   NULL,
   NULL,
   NULL,
   NULL,
   &def_size,
   &def_size,
   &def_depth,
   NULL,
   NULL,
   FALSE,
   FALSE,
   NULL,
   NULL,
   FALSE,
   FALSE,
   "56K",
   FALSE,
   NULL,
   FALSE,
   FALSE,
   FALSE,
   NULL,
   NULL
};

static struct WBStartup*  wb_msg = NULL;
int a_slowmouse;

static void rdesktop_reset_state(void)
{
  rdp_reset_state();
}

static RD_BOOL read_password(char *password, int size)
{
  RD_BOOL ret = False;  
  char *p;

  fprintf(stderr, "Password: \e[8m");

  if (fgets(password, size, stdin) != NULL)
  {
    ret = True;

    /* strip final newline */
    p = strchr(password, '\n');
    if (p != NULL)
      *p = 0;
  }

  fprintf(stderr, "\e[28m\n");

  return ret;
}

static int 
parse_server_and_port(char *server)
{
  char *p;
  int port = 0;

  p = strchr(server, ':');
  if (p != NULL)
  {
   port = g_tcp_port_rdp = strtol(p + 1, NULL, 10);
    *p = 0;
  }
  return port;
}

enum { UI_INIT          = 1,
       RDPSND_INIT      = 2,
       RDPDR_INIT       = 4,
       RDP_CONNECT      = 8,
       UI_CREATE_WINDOW = 16};

static int startup = 0;

static void cleanup(void)
{
  if (ext_disc_reason >= 2) 
  {
    show_disconnect_reason(ext_disc_reason);
  }

  if( amiga_icon != NULL )
  {
    FreeDiskObject( amiga_icon );
  }

  if (startup & RDPSND_INIT) {
#ifdef WITH_RDPSND
     rdpsnd_deinit();
#endif
  }

  if (startup & UI_CREATE_WINDOW) 
  {
    extern int g_pstcache_fd[8];
    int i;

    cache_save_state();
    ui_destroy_window();
    for (i = 0 ; i < 8 ; i++)
    {
      if (g_pstcache_fd[i])
      {
        Close(g_pstcache_fd[i]);
        g_pstcache_fd[i]=0;
      }
    }
  }

  if (startup & RDP_CONNECT) {  
    DEBUG(("Disconnecting...\n"));
    rdp_disconnect();
  }

  if (startup & UI_INIT) {
    ui_deinit();
  }
}


/* Client program */
int main(int argc, char *argv[])
{
	char server[64];
	char net_disk[1024];
	char fullhostname[64];
	char domain[16];
	char password[64];
	char shell[256];
	char directory[256];
	RD_BOOL prompt_password, deactivated;
	struct passwd *pw;
	uint32 flags;
	int port;
	char *p;


	char pubscreen[64];
	struct RDArgs* rdargs = NULL;
	char* wbargs = NULL;
	int rc = RETURN_OK;

	int run_count = 0;	/* Session Directory support */
	BOOL continue_connect = True;	/* Session Directory support */

	atexit(cleanup);

	// Don't care if we're unable to open it

	flags = RDP_LOGON_NORMAL;
	prompt_password = False;
	domain[0] = password[0] = shell[0] = directory[0] = 0;
	strcpy(g_keymapname, "en-us");
	g_embed_wnd = 0;


	pw = getpwuid(getuid());
	if((pw != NULL) && (pw->pw_name != NULL))
	{
		a_args.a_user = pw->pw_name;
	}

	if(gethostname(fullhostname, sizeof(fullhostname)) != -1)
	{
		p = strchr(fullhostname, '.');
		if(p != NULL)
			*p = 0;

		a_args.a_client = fullhostname;
	}

	g_keylayout = getkeylayout(g_keylayout);

	if(argc == 0)
	{
		struct DiskObject* icon;
		BPTR cd;
		int length = 1; // space for one nul character
		int i, t;

		cd = CurrentDir(0);

		wb_msg = (struct WBStartup*) argv;

		for(i = wb_msg->sm_NumArgs - 1; i >= 0; --i)
		{
			CurrentDir(wb_msg->sm_ArgList[ i ].wa_Lock);

			icon = GetDiskObject(wb_msg->sm_ArgList[ i ].wa_Name);

			if(amiga_icon == NULL)
			{
				amiga_icon = icon;
			}

			if(icon != NULL && icon->do_ToolTypes != NULL)
			{
				for(t = 0;  icon->do_ToolTypes[ t ] != NULL; ++t)
				{
					char* tt = icon->do_ToolTypes[ t ];

					if(isalpha(tt[ 0 ]) && strncmp(tt, "WINDOW", 6) != 0)
					{
						// Add length + space
						length = length + strlen(tt) + 1;
					}
				}

				if(icon != amiga_icon)
				{
					FreeDiskObject(icon);
				}
			}
		}

		wbargs = AllocVec(length, MEMF_ANY);
		wbargs[0] = 0;

		if(wbargs != NULL)
		{
			for(i = wb_msg->sm_NumArgs - 1; i >= 0; --i)
			{
				CurrentDir(wb_msg->sm_ArgList[ i ].wa_Lock);

				icon = GetDiskObject(wb_msg->sm_ArgList[ i ].wa_Name);

				if(icon != NULL && icon->do_ToolTypes != NULL)
				{
					for(t = 0;  icon->do_ToolTypes[ t ] != NULL; ++t)
					{
						char* tt = icon->do_ToolTypes[ t ];

						if(isalpha(tt[ 0 ]) && Strnicmp(tt, "WINDOW", 6) != 0)
						{
							int tt_len;
							char* args = wbargs;

							for(tt_len = 0; isalnum(tt[ tt_len ]); ++tt_len);

							while(1)
							{
								if(Strnicmp(args, tt, tt_len) == 0 &&
								        ! isalnum(tt[ tt_len ]))
								{
									break;
								}
								else
								{
									while(*args != 0)
									{
										++args;

										if(*args == '\n')
										{
											++args;
											break;
										}
									}

									if(*args == 0)
									{
										strcat(args, tt);
										strcat(args, "\n");   // Marker!
										break;
									}
								}
							}
						}
					}

					FreeDiskObject(icon);
				}
			}

			// Fix for ReadArgs: replace all but the last LF with a space
			length = strlen(wbargs);

			for(i = 0; i < length - 1; ++i)
			{
				if(wbargs[ i ] == '\n')
				{
					wbargs[ i ] = ' ';
				}
			}

			rdargs = AllocDosObjectTags(DOS_RDARGS, TAG_DONE);

			if(rdargs != NULL)
			{
				rdargs->RDA_Source.CS_Buffer = wbargs;
				rdargs->RDA_Source.CS_Length = length;
				rdargs->RDA_Source.CS_CurChr = 0;
			}
		}

		CurrentDir(cd);
	}

	if(amiga_icon == NULL)
	{
		BPTR cd = CurrentDir(GetProgramDir());
		char name[ 256 ];

		if(GetProgramName(name, sizeof(name)))
		{
			amiga_icon = GetDiskObjectNew(name[0] ? name : "RDesktop");
		}

		CurrentDir(cd);
	}

	if(0 == argc)
	{
		rdargs = ReadArgs(wbtemplate, (LONG*) &a_args, rdargs);
	}
	else
	{
		rdargs = ReadArgs(template, (LONG*)&a_args, rdargs);
	}

	if(rdargs == NULL)
	{
		char txt[128];

		Fault(IoErr(), NULL, txt, sizeof(txt));
		error("%s\n", txt);
		return RETURN_ERROR;
	}


	if(a_args.a_user == NULL)
	{
		error("Could not determine username.\nYou have to specify it manually.\n");
		rc = RETURN_ERROR;
	}

	if(a_args.a_client == NULL)
	{
		error("Could not determine client name.\nYou have to specify it manually.\n");
		rc = RETURN_ERROR;
	}

	if(a_args.a_server != NULL) STRNCPY(server, a_args.a_server, sizeof(server));
	if(a_args.a_user != NULL)
	{
		STRNCPY(g_username, a_args.a_user, sizeof(g_username));
	}
	if(a_args.a_domain != NULL) STRNCPY(domain, a_args.a_domain, sizeof(domain));
	if(a_args.a_shell != NULL)  STRNCPY(shell, a_args.a_shell, sizeof(shell));
	if(a_args.a_dir != NULL)    STRNCPY(directory, a_args.a_dir, sizeof(directory));
	if(a_args.a_password != NULL)
	{
		STRNCPY(password, a_args.a_password, sizeof(password));
		flags |= RDP_LOGON_AUTO;
	}
	if(a_args.a_client != NULL) STRNCPY(g_hostname, a_args.a_client, sizeof(g_hostname));

	if(a_args.a_pubscreen != NULL)
	{
		STRNCPY(pubscreen, a_args.a_pubscreen, sizeof(pubscreen));
		amiga_pubscreen_name = pubscreen;
	}

	if(a_args.a_screenmode != NULL)
	{
		amiga_screen_id = *a_args.a_screenmode;
		a_args.a_fullscreen = TRUE;
	}

	g_fullscreen      = a_args.a_fullscreen;
	g_clonescreen     = a_args.a_clonescreen;
	if(a_args.a_left != NULL) amiga_left = *a_args.a_left;
	if(a_args.a_top != NULL)  amiga_top  = *a_args.a_top;

	if(a_args.a_fullscreen && (a_args.a_left != NULL || a_args.a_top != NULL || a_args.a_pubscreen != NULL))
	{
		warning("The LEFT, TOP and PUBSCREEN arguments are ignored in fullscreen mode.\n");
	}

	g_width             = *a_args.a_width;
	g_height            = *a_args.a_height;

	g_encryption        = ! a_args.a_french;
	g_packet_encryption = ! a_args.a_noenc;
	g_sendmotion        = ! a_args.a_nomouse;
#ifdef WITH_RDPSND
	if(a_args.a_audio != NULL) amiga_audio_unit = *a_args.a_audio;
#endif
	if(a_args.a_remoteaudio)
	{
		flags |= RDP_LOGON_LEAVE_AUDIO;

		if(a_args.a_audio != NULL)
		{
			warning("The AUDIO argument is ignored when REMOTEAUDIO is used.\n");
		}
	}
#ifdef WITH_RDPSND
	else
	{
		g_rdpsnd = True;
	}
#endif

	if(a_args.a_title != NULL) STRNCPY(g_title, a_args.a_title, sizeof(g_title));

	g_server_depth = *a_args.a_depth;

	if(g_server_depth < 0 ||
	        (g_server_depth > 8 && g_server_depth != 16 && g_server_depth != 15
	         && g_server_depth != 24))
	{
		error("invalid server colour depth.\n");
		rc = RETURN_ERROR;
	}

	if(a_args.a_experience != NULL)
	{
		toupper_str(a_args.a_experience);

		if(strcmp("28K8", a_args.a_experience) == 0)
		{
			g_rdp5_performanceflags =
			    RDP5_NO_WALLPAPER | RDP5_NO_FULLWINDOWDRAG |
			    RDP5_NO_MENUANIMATIONS | RDP5_NO_THEMING;
		}
		else if(strcmp("56K", a_args.a_experience) == 0)
		{
			g_rdp5_performanceflags =
			    RDP5_NO_WALLPAPER | RDP5_NO_FULLWINDOWDRAG |
			    RDP5_NO_MENUANIMATIONS;
		}
		else if(strcmp("DSL", a_args.a_experience) == 0)
		{
			g_rdp5_performanceflags = RDP5_NO_WALLPAPER;
		}
		else if(strcmp("LAN", a_args.a_experience) == 0)
		{
			g_rdp5_performanceflags = RDP5_DISABLE_NOTHING;
		}
		else
		{
			error("Illegal EXPERIENCE value: %s\n", a_args.a_experience);
			rc = RETURN_ERROR;
		}
	}

	if(a_args.a_netdisk)
	{
		strcpy(net_disk, "netdisk=");
		strcat(net_disk, a_args.a_netdisk);

		g_num_devices = disk_enum_devices(&g_num_devices, net_disk);
	}
	g_console_session = a_args.a_console;
	g_use_rdp5 = ! a_args.a_rdp4;

	if(a_args.a_keymap != NULL) g_keylayout = *a_args.a_keymap;

	if(a_args.a_rdp_compression) flags |= (RDP_LOGON_COMPRESSION | RDP_LOGON_COMPRESSION2);
	g_bitmap_cache_persist_enable = 0;
	g_bitmap_cache_persist_enable = a_args.a_persist_cache;
	if(g_sendmotion && a_args.a_slowmouse)
	{
		a_slowmouse = True;
		g_sendmotion = False;
	}

	if(server[0] == 0)
	{
		error("No server specified\n");
		rc = RETURN_ERROR;
	}

	port = parse_server_and_port(server);

	if(strcmp(password, "-") == 0)
	{
		read_password(password, sizeof(password));
	}

	if(g_title[0] == 0)
	{
		if(port)
			sprintf(g_title,"RDesktop - %s:%d", server, port);
		else sprintf(g_title,"RDesktop - %s", server);
	}

	FreeArgs(rdargs);
	FreeVec(wbargs);

	// We don't support this at the moment
	g_polygon_ellipse_orders = False;

#define PRINTS( s ) //printf( "%s: %s\n", #s, s );
#define PRINTI( i ) //printf( "%s: %d\n", #i, i );

	PRINTI(amiga_left);
	PRINTI(amiga_top);
	PRINTI(amiga_screen_id);
	PRINTS(amiga_pubscreen_name);

	PRINTS(g_title);
	PRINTS(g_username);
	PRINTS(g_hostname);
	PRINTS(g_keymapname);
	PRINTI(g_keylayout);
	PRINTI(g_width);
	PRINTI(g_height);
	PRINTI(tcp_port_rdp);
	PRINTI(g_server_depth);
	PRINTI(g_win_button_size);
	PRINTI(g_bitmap_compression);
	PRINTI(g_sendmotion);
	PRINTI(g_encryption);
	PRINTI(g_packet_encryption);
	PRINTI(g_desktop_save);
	PRINTI(g_fullscreen);
	PRINTI(g_grab_keyboard);
	PRINTI(g_hide_decorations);
	PRINTI(g_use_rdp5);
	PRINTI(g_console_session);
	PRINTI(g_numlock_sync);
	PRINTI(g_owncolmap);
	PRINTI(g_ownbackstore);
	PRINTI(g_embed_wnd);
	PRINTI(g_rdp5_performanceflags);

	if(rc != RETURN_OK)
	{
		return rc;
	}

	if(!ui_init())
		return RETURN_FAIL;

	startup |= UI_INIT;

#ifdef WITH_RDPSND
	if(g_rdpsnd && rdpsnd_init(NULL))
	{
		startup |= RDPSND_INIT;	
	}
#endif
	rdpdr_init();
	startup |= RDPDR_INIT;

	while(1)	/* add support for Session Directory; only reconnect once */
	{
		rdesktop_reset_state();

		if(g_redirect)
		{
			STRNCPY(domain, g_redirect_domain, sizeof(domain));
			STRNCPY(g_username, g_redirect_username, strlen(g_redirect_username) + 1);
			STRNCPY(password, g_redirect_password, sizeof(password));
			STRNCPY(server, g_redirect_server, sizeof(server));
			flags |= RDP_LOGON_AUTO;
		}


		if(!rdp_connect(server, flags, domain, password, shell, directory, g_redirect))
			return EX_PROTOCOL;

		startup |= RDP_CONNECT;
		/* By setting encryption to False here, we have an encrypted login
		   packet but unencrypted transfer of other packets */
		if(!g_packet_encryption)
			g_encryption = False;

		DEBUG(("Connection successful.\n"));
		memset(password, 0, sizeof(password));

		if(!g_redirect)
			if(!ui_create_window())
			{
				return EX_OSERR;
			}
			else
			{
				startup |= UI_CREATE_WINDOW;
			}
		g_redirect = False;
		rdp_main_loop(&deactivated, &ext_disc_reason);

		rdp_disconnect();    
		startup &= ~RDP_CONNECT;


		/*if (g_redirect)
			continue;
		*/
#ifndef __MORPHOS__
		if(g_pending_resize)
		{
			/* If we have a pending resize, reconnect using the new size, rather than exit */
			g_pending_resize = False;
			continue;
		}
#endif
		break;
	}

	if(deactivated)
		return RETURN_OK;
	else if(ext_disc_reason == exDiscReasonAPIInitiatedDisconnect
	        || ext_disc_reason == exDiscReasonAPIInitiatedLogoff)
		return RETURN_OK;
	else
		return RETURN_WARN;
}

void amiga_req(char* prefix, char* txt);

static void showmsg(char* prefix, char* format, va_list ap)
{
  char txt[256];

  vsnprintf(txt, sizeof(txt), format, ap);
  va_end(ap);

  if( wb_msg == NULL || IntuitionBase == NULL )
  {
    fprintf(stderr, "%s: %s", prefix, txt);
  }
  else
  {
    int lf = strlen( txt ) - 1;

    if( lf >= 0 &&  txt[ lf ] == '\n' )
    {
      txt[ lf ] = '\0';
    }
    
    amiga_req( prefix, txt );
  }
}

/* report an error */
void error(char *format, ...)
{
  va_list ap;
  va_start(ap, format);

  showmsg( "ERROR", format, ap );
  va_end(ap);
}

/* report a warning */
void warning(char *format, ...)
{
  va_list ap;
  va_start(ap, format);

  showmsg( "WARNING", format, ap );
  va_end(ap);
}

/* report an unimplemented protocol feature */
void unimpl(char *format, ...)
{
  va_list ap;
  va_start(ap, format);

  showmsg( "NOT IMPLEMENTED", format, ap );
  va_end(ap);
}

static void show_disconnect_reason(uint16 reason)
{
	char *text;

	switch (reason)
	{
		case exDiscReasonNoInfo:
			text = "No information available";
			break;

		case exDiscReasonAPIInitiatedDisconnect:
			text = "Server initiated disconnect";
			break;

		case exDiscReasonAPIInitiatedLogoff:
			text = "Server initiated logoff";
			break;

		case exDiscReasonServerIdleTimeout:
			text = "Server idle timeout reached";
			break;

		case exDiscReasonServerLogonTimeout:
			text = "Server logon timeout reached";
			break;

		case exDiscReasonReplacedByOtherConnection:
			text = "The session was replaced";
			break;

		case exDiscReasonOutOfMemory:
			text = "The server is out of memory";
			break;

		case exDiscReasonServerDeniedConnection:
			text = "The server denied the connection";
			break;

		case exDiscReasonServerDeniedConnectionFips:
			text = "The server denied the connection for security reason";
			break;

		case exDiscReasonLicenseInternal:
			text = "Internal licensing error";
			break;

		case exDiscReasonLicenseNoLicenseServer:
			text = "No license server available";
			break;

		case exDiscReasonLicenseNoLicense:
			text = "No valid license available";
			break;

		case exDiscReasonLicenseErrClientMsg:
			text = "Invalid licensing message";
			break;

		case exDiscReasonLicenseHwidDoesntMatchLicense:
			text = "Hardware id doesn't match software license";
			break;

		case exDiscReasonLicenseErrClientLicense:
			text = "Client license error";
			break;

		case exDiscReasonLicenseCantFinishProtocol:
			text = "Network error during licensing protocol";
			break;

		case exDiscReasonLicenseClientEndedProtocol:
			text = "Licensing protocol was not completed";
			break;

		case exDiscReasonLicenseErrClientEncryption:
			text = "Incorrect client license enryption";
			break;

		case exDiscReasonLicenseCantUpgradeLicense:
			text = "Can't upgrade license";
			break;

		case exDiscReasonLicenseNoRemoteConnections:
			text = "The server is not licensed to accept remote connections";
			break;

		default:
			if (reason > 0x1000 && reason < 0x7fff)
			{
				text = "Internal protocol error";
			}
			else
			{
				text = "Unknown reason";
			}
	}
	error("disconnect: %s.\n", text);
}
